////////////////////////////////////////////////////////////////////////////
//
// nrModel.cpp
//
// Nate Robins, November 2001.
//
// Model class.
//
////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////
// Includes
////////////////////////////////////////////////////////////////////////////

#include "nrModel.h"

#include "nrVector2.h"
#include "nrVector3.h"

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <string.h>


////////////////////////////////////////////////////////////////////////////
// Implementation
////////////////////////////////////////////////////////////////////////////

// nrTriangle //////////////////////////////////////////////////////////////

nrmTriangle::nrmTriangle()
{
    m_Model = NULL;
    m_Vertices[ 0 ].m_VectorIndex   = 0;
    m_Vertices[ 0 ].m_NormalIndex   = 0;
    m_Vertices[ 0 ].m_TexCoordIndex = 0;
    m_Vertices[ 1 ].m_VectorIndex   = 0;
    m_Vertices[ 1 ].m_NormalIndex   = 0;
    m_Vertices[ 1 ].m_TexCoordIndex = 0;
    m_Vertices[ 2 ].m_VectorIndex   = 0;
    m_Vertices[ 2 ].m_NormalIndex   = 0;
    m_Vertices[ 2 ].m_TexCoordIndex = 0;
    m_FacetNormalIndex = 0;
}

////////////////////////////////////////////////////////////////////////////

nrmTriangle::~nrmTriangle()
{
}

////////////////////////////////////////////////////////////////////////////

const nrVector3& nrmTriangle::Vector( int n ) const
{
    assert( n >= 0 );
    assert( n < 3 );
    assert( m_Model != NULL );
    
    int i = m_Vertices[ n ].m_VectorIndex;
    
    assert( i >= 0 );
    assert( i < m_Model->m_NumVectors );
    
    return m_Model->m_Vectors[ i ];
}

////////////////////////////////////////////////////////////////////////////

const nrVector3& nrmTriangle::Normal( int n ) const
{
    assert( n >= 0 );
    assert( n < 3 );
    assert( m_Model != NULL );
    
    int i = m_Vertices[ n ].m_NormalIndex;
    
    assert( i >= 0 );
    assert( i < m_Model->m_NumNormals );
    
    return m_Model->m_Normals[ i ];
}

////////////////////////////////////////////////////////////////////////////

const nrColor& nrmTriangle::Color( int n ) const
{
    assert( n >= 0 );
    assert( n < 3 );
    assert( m_Model != NULL );
    
    int i = m_Vertices[ n ].m_ColorIndex;
    
    assert( i >= 0 );
    assert( i < m_Model->m_NumColors );
    
    return m_Model->m_Colors[ i ];
}

////////////////////////////////////////////////////////////////////////////

const nrVector2& nrmTriangle::TexCoord( int n ) const
{
    assert( n >= 0 );
    assert( n < 3 );
    assert( m_Model != NULL );
    
    int i = m_Vertices[ n ].m_TexCoordIndex;
    
    assert( i >= 0 );
    assert( i < m_Model->m_NumTexCoords );
    
    return m_Model->m_TexCoords[ i ];
}

////////////////////////////////////////////////////////////////////////////

const nrVector3& nrmTriangle::FacetNormal( void ) const
{
    assert( m_Model != NULL );
    
    int i = m_FacetNormalIndex;
    
    assert( i >= 0 );
    assert( i < m_Model->m_NumFacetNormals );
    
    return m_Model->m_FacetNormals[ i ];
}

////////////////////////////////////////////////////////////////////////////
// nrMaterial
////////////////////////////////////////////////////////////////////////////

nrmMaterial::nrmMaterial()
{
    m_Name = NULL;
    m_Ambient = nrColor( 0.2f, 0.2f, 0.2f );
    m_Diffuse = nrColor( 0.8f, 0.8f, 0.8f );
    m_Specular = nrColor( 0.0f, 0.0f, 0.0f );
    m_Emissive = nrColor( 0.0f, 0.0f, 0.0f );
    m_Shininess = 0.0f;
}

////////////////////////////////////////////////////////////////////////////

nrmMaterial::~nrmMaterial()
{
    delete [] m_Name;
}

////////////////////////////////////////////////////////////////////////////
// nrGroup
////////////////////////////////////////////////////////////////////////////

nrmGroup::nrmGroup()
{
    m_Name = NULL;
    m_Model = NULL;
    m_NumTriangleIndices = 0;
    m_TriangleIndices = NULL;
    m_MaterialIndex = 0;
}

////////////////////////////////////////////////////////////////////////////

nrmGroup::~nrmGroup()
{
    delete [] m_Name;
    delete [] m_TriangleIndices;
}

////////////////////////////////////////////////////////////////////////////

const char* nrmGroup::Name( void ) const
{
    return m_Name;
}

////////////////////////////////////////////////////////////////////////////

void nrmGroup::SetName( const char* name )
{
    assert( name );
    
    delete [] m_Name;
    m_Name = new char[ strlen( name ) + 1 ];
    if ( m_Name == NULL )
    {
        printf( "nrGroup::SetName(): Out of memory.\n" );
    }
    else
    {
        strcpy( m_Name, name );
    }
}

////////////////////////////////////////////////////////////////////////////

void nrmGroup::SetMaterial( int index )
{
    m_MaterialIndex = index;
}

////////////////////////////////////////////////////////////////////////////

int nrmGroup::NumTriangles( void ) const
{
    return m_NumTriangleIndices;
}

////////////////////////////////////////////////////////////////////////////

const nrmTriangle& nrmGroup::Triangle( int n ) const
{
    assert( n >= 0 );
    assert( n < m_NumTriangleIndices );
    assert( m_Model != NULL );
    
    int i = m_TriangleIndices[ n ];
    
    assert( i >= 0 );
    assert( i < m_Model->m_NumTriangles );
    
    return m_Model->m_Triangles[ i ];
}

////////////////////////////////////////////////////////////////////////////

const nrmMaterial& nrmGroup::Material( void ) const
{
    assert( m_Model != NULL );
    
    int i = m_MaterialIndex;
    
    assert( i >= 0 );
    assert( i < m_Model->m_NumMaterials );
    
    return m_Model->m_Materials[ i ];
}

////////////////////////////////////////////////////////////////////////////
// nrModel
////////////////////////////////////////////////////////////////////////////

nrModel::nrModel()
{
    m_Name = NULL;
    m_Path = NULL;
    
    m_NumVectors = 0;
    m_Vectors    = NULL;
    
    m_NumColors = 0;
    m_Colors    = NULL;
    
    m_NumNormals = 0;
    m_Normals    = NULL;
    
    m_NumTexCoords = 0;
    m_TexCoords    = NULL;
    
    m_NumFacetNormals = 0;
    m_FacetNormals    = NULL;
    
    m_NumTriangles = 0;
    m_Triangles    = NULL;
    
    m_NumMaterials = 0;
    m_Materials    = NULL;
    
    m_NumGroups = 0;
    m_Groups    = NULL;
}

////////////////////////////////////////////////////////////////////////////

nrModel::~nrModel()
{
    delete [] m_Name;
    delete [] m_Path;
    delete [] m_Vectors;
    delete [] m_Colors;
    delete [] m_Normals;
    delete [] m_TexCoords;
    delete [] m_FacetNormals;
    delete [] m_Triangles;
    delete [] m_Materials;
    delete [] m_Groups;
}

////////////////////////////////////////////////////////////////////////////

int nrModel::NumVectors( void ) const
{
    return m_NumVectors;
}

////////////////////////////////////////////////////////////////////////////

const nrVector3& nrModel::Vector( int n ) const
{
    assert( n >= 0 && n < m_NumVectors );
    
    return m_Vectors[ n ];
}

////////////////////////////////////////////////////////////////////////////

int nrModel::NumNormals( void ) const
{
    return m_NumNormals;
}

////////////////////////////////////////////////////////////////////////////

const nrVector3& nrModel::Normal( int n ) const
{
    assert( n >= 0 && n < m_NumNormals );
    
    return m_Normals[ n ];
}

////////////////////////////////////////////////////////////////////////////

int nrModel::NumColors( void ) const
{
    return m_NumColors;
}

////////////////////////////////////////////////////////////////////////////

const nrColor& nrModel::Color( int n ) const
{
    assert( n >= 0 && n < m_NumColors );
    
    return m_Colors[ n ];
}

////////////////////////////////////////////////////////////////////////////

int nrModel::NumTexCoords( void ) const
{
    return m_NumTexCoords;
}

////////////////////////////////////////////////////////////////////////////

const nrVector2& nrModel::TexCoord( int n ) const
{
    assert( n >= 0 && n < m_NumTexCoords );
    
    return m_TexCoords[ n ];
}

////////////////////////////////////////////////////////////////////////////

int nrModel::NumFacetNormals( void ) const
{
    return m_NumFacetNormals;
}

////////////////////////////////////////////////////////////////////////////

const nrVector3& nrModel::FacetNormal( int n ) const
{
    assert( n >= 0 && n < m_NumFacetNormals );
    
    return m_FacetNormals[ n ];
}

////////////////////////////////////////////////////////////////////////////

int nrModel::NumTriangles( void ) const
{
    return m_NumTriangles;
}

////////////////////////////////////////////////////////////////////////////

const nrmTriangle& nrModel::Triangle( int n ) const
{
    assert( n >= 0 && n < m_NumTriangles );
    
    return m_Triangles[ n ];
}

////////////////////////////////////////////////////////////////////////////

int nrModel::NumMaterials( void ) const
{
    return m_NumMaterials;
}

////////////////////////////////////////////////////////////////////////////

const nrmMaterial& nrModel::Material( int n ) const
{
    assert( n >= 0 && n < m_NumMaterials );
    
    return m_Materials[ n ];
}

////////////////////////////////////////////////////////////////////////////

int nrModel::NumGroups( void ) const
{
    return m_NumGroups;
}

////////////////////////////////////////////////////////////////////////////

const nrmGroup& nrModel::Group( int n ) const
{
    assert( n >= 0 && n < m_NumGroups );
    
    return m_Groups[ n ];
}

////////////////////////////////////////////////////////////////////////////

void nrModel::Dimensions( nrVector3& mins, nrVector3& maxs ) const
{
    mins.x = 2e31f;
    mins.y = 2e31f;
    mins.z = 2e31f;
    
    maxs.x = -2e31f;
    maxs.y = -2e31f;
    maxs.z = -2e31f;
    
    for ( int i = 0; i < m_NumVectors; i++ )
    {
        const nrVector3& v = Vector( i );
        
        if ( v.x < mins.x )
        {
            mins.x = v.x;
        }
        if ( v.x > maxs.x )
        {
            maxs.x = v.x;
        }
        
        if ( v.y < mins.y )
        {
            mins.y = v.y;
        }
        if ( v.y > maxs.y )
        {
            maxs.y = v.y;
        }
        
        if ( v.z < mins.z )
        {
            mins.z = v.z;
        }
        if ( v.z > maxs.z )
        {
            maxs.z = v.z;
        }
    }
}

////////////////////////////////////////////////////////////////////////////

void nrModel::RenderGroup( int n, unsigned int flags ) const
{
#if 0
    assert( n >= 0 );
    assert( n < m_NumGroups );
    
    const nrGroup& group = Group( n );
    
    if ( flags & RENDER_FACETNORMALS && m_FacetNormals == NULL )
    {
        printf( "nrModel::RenderGroup(): RENDER_FACETNORMALS flag set without any facet normals (ignored).\n" );
        flags &= ~RENDER_FACETNORMALS;
    }
    if ( flags & RENDER_NORMALS && m_Normals == NULL )
    {
        printf( "nrModel::RenderGroup(): RENDER_NORMALS flag set without any normals (ignored).\n" );
        flags &= ~RENDER_NORMALS;
    }
    if ( flags & RENDER_TEXCOORDS && m_TexCoords == NULL )
    {
        printf( "nrModel::RenderGroup(): RENDER_TEXCOORDS flag set without any texcoords (ignored).\n" );
        flags &= ~RENDER_TEXCOORDS;
    }
    if ( flags & RENDER_COLORS && m_Colors == NULL )
    {
        printf( "nrModel::RenderGroup(): RENDER_COLORS flag set without any colors (ignored).\n" );
        flags &= ~RENDER_COLORS;
    }
    
    if ( flags & RENDER_COLORS )
    {
        glEnable( GL_COLOR_MATERIAL );
    }
    else
    {
        glDisable( GL_COLOR_MATERIAL );
    }
    
    glBegin( GL_TRIANGLES );
    
    for ( int i = 0; i < group.NumTriangles(); i++ )
    {
        const nrTriangle& t = group.Triangle( i );
        
        for ( int v = 0; v < 3; v++ )
        {
            if ( flags & RENDER_FACETNORMALS )
            {
                glNormal3fv( ( float* )&( t.FacetNormal() ) );
            }
            if ( flags & RENDER_NORMALS )
            {
                glNormal3fv( ( float* )&( t.Normal( v ) ) );
            }
            if ( flags & RENDER_COLORS )
            {
                glColor3fv( ( float* )&( t.Color( v ) ) );
            }
            glVertex3fv( ( float* )&( t.Vector( v ) ) );
        }
    }
    
    glEnd();
#endif
}

////////////////////////////////////////////////////////////////////////////

void nrModel::Render( unsigned int flags ) const
{
    for ( int i = 0; i < NumGroups(); i++ )
    {
        RenderGroup( i, flags );
    }
}

////////////////////////////////////////////////////////////////////////////

void nrModel::CreateFacetNormals( void )
{
    // Clobber any old facet normals.
    m_NumFacetNormals = 0;
    delete [] m_FacetNormals;
    
    // Allocate memory for the new facet normals.
    m_NumFacetNormals = m_NumTriangles;
    m_FacetNormals = new nrVector3[ m_NumFacetNormals ];
    if ( m_FacetNormals == NULL )
    {
        printf( "nrModel::CreateFacetNormals(): Out of memory.\n" );
        return;
    }
    
    for ( int i = 0; i < m_NumTriangles; i++ )
    {
        nrmTriangle& t = m_Triangles[ i ];
        
        t.m_FacetNormalIndex = i;
        
        const nrVector3& v0 = t.Vector( 0 );
        const nrVector3& v1 = t.Vector( 1 );
        const nrVector3& v2 = t.Vector( 2 );
        
        nrVector3 u = v1 - v0;
        nrVector3 v = v2 - v0;
        nrVector3 n = u.Cross( v ).Unit();
        
        m_FacetNormals[ i ] = n;
    }
}

////////////////////////////////////////////////////////////////////////////

void nrModel::CreateNormals( void )
{
    printf( "Creating normals for model \"%s\".\n", m_Name );
    
    // Clobber any old normals.
    m_NumNormals = 0;
    delete [] m_Normals;
    
    // Must have facet normals to create vertex normals.
    if ( m_NumFacetNormals == 0 )
    {
        CreateFacetNormals();
    }
    
    // Allocate memory for the new vertex normals.
    m_NumNormals = m_NumVectors;
    m_Normals = new nrVector3[ m_NumNormals ];
    if ( m_Normals == NULL )
    {
        printf( "nrModel::CreateNormals(): Out of memory.\n" );
        return;
    }
    
    int total = m_NumTriangles + m_NumVectors + m_NumVectors + m_NumTriangles;
    int current = 0;
    
    // Allocate memory for a structure to speed up the search for the 
    // triangles that a vector belongs to.  (record all of the triangles 
    // that each vector is in).
    typedef struct 
    {
        int num_triangles;
        int total_triangles;
        int* triangles;
    } optimize;
    
    optimize* speed = new optimize[ m_NumVectors ];
    memset( speed, 0, sizeof( optimize ) * m_NumVectors );
    
    // Populate the optimized structure.
    int i;
    for ( i = 0; i < m_NumTriangles; i++ )
    {
        nrmTriangle& t = m_Triangles[ i ];
        
        for ( int j = 0; j < 3; j++ )
        {
            optimize& s = speed[ t.m_Vertices[ j ].m_VectorIndex ];
            
            if ( s.num_triangles + 1 > s.total_triangles )
            {
                int* triangles = s.triangles;
                
                s.total_triangles += 5;
                s.triangles = new int[ s.total_triangles ];
                if ( s.triangles == NULL )
                {
                    printf( "nrModel::CreateNormals(): Out of memory.\n" );
                    return;
                }
                memcpy( s.triangles, triangles, s.num_triangles * sizeof ( int ) );
                delete [] triangles;
                
                s.triangles[ s.num_triangles ] = i;
                s.num_triangles++;
            }
            else
            {
                s.triangles[ s.num_triangles ] = i;
                s.num_triangles++;
            }
        }
        
        if ( current % 100 == 0 )
        {
            Working( ( float )current / ( float )total );
        }
        
        current++;
    }
    
    // Average normals.
    for ( i = 0; i < m_NumVectors; i++ )
    {
        nrVector3 n = nrVector3( 0.0f, 0.0f, 0.0f );
        optimize& s = speed[ i ];
        
        for ( int j = 0; j < s.num_triangles; j++ )
        {
            nrmTriangle& t = m_Triangles[ s.triangles[ j ] ];
            nrVector3&   fn = m_FacetNormals[ t.m_FacetNormalIndex ];
            
            n = n + fn;
        }
        
        n = n.Unit();
        
        m_Normals[ i ] = n;
        
        if ( current % 100 == 0 )
        {
            Working( ( float )current / ( float )total );
        }
        
        current++;
    }
    
    // Delete the optimized structure
    for ( i = 0; i < m_NumVectors; i++ )
    {
        optimize& s = speed[ i ];
        
        delete [] s.triangles;
        
        if ( current % 100 == 0 )
        {
            Working( ( float )current / ( float )total );
        }
        
        current++;
    }
    delete [] speed;
    
    // This is the old, elegant, but slow (no, really slow) way, which 
    // the above supercedes.
    /*
    for ( int i = 0; i < m_NumVectors; i++ )
    {
    nrVector3 n = nrVector3( 0.0f, 0.0f, 0.0f );
    
      for ( int j = 0; j < m_NumTriangles; j++ )
      {
      nrTriangle& t  = m_Triangles[ j ];
      nrVector3&   fn = m_FacetNormals[ t.m_FacetNormalIndex ];
      
        if ( t.m_Vertices[ 0 ].m_VectorIndex == i ||
        t.m_Vertices[ 1 ].m_VectorIndex == i ||
        t.m_Vertices[ 2 ].m_VectorIndex == i )
        {
        n = n + fn;
        }
        }
        
          n.Normalize();
          
            m_Normals[ i ] = n;
            
              if ( current % 100 == 0 )
              {
              Working( ( float )current / ( float )total );
              }
              
                current++;
                }
    */
    
    // Re-index the normals.
    for( i = 0; i < m_NumTriangles; i++ )
    {
        nrmTriangle& t = m_Triangles[ i ];
        
        t.m_Vertices[ 0 ].m_NormalIndex = t.m_Vertices[ 0 ].m_VectorIndex;
        t.m_Vertices[ 1 ].m_NormalIndex = t.m_Vertices[ 1 ].m_VectorIndex;
        t.m_Vertices[ 2 ].m_NormalIndex = t.m_Vertices[ 2 ].m_VectorIndex;
        
        if ( current % 100 == 0 )
        {
            Working( ( float )current / ( float )total );
        }
        
        current++;
    }
    
    Working( 1.0f );
    printf( "\nFinished.\n" );
}

////////////////////////////////////////////////////////////////////////////

void nrModel::CreateColors( void )
{
    printf( "Creating colors for model \"%s\".\n", m_Name );
    
    // Clobber any old colors.
    m_NumColors = 0;
    delete [] m_Colors;
    
    // Must have normals to create colors.
    if ( m_NumNormals == 0 )
    {
        CreateNormals();
    }
    
    // Allocate memory for the new colors.
    m_NumColors = m_NumVectors;
    m_Colors = new nrColor[ m_NumColors ];
    if ( m_Colors == NULL )
    {
        printf( "nrModel::CreateColors(): Out of memory.\n" );
        return;
    }
    
    int total = m_NumColors + m_NumTriangles;
    int current = 0;
    
    int i;
    for ( i = 0; i < m_NumColors; i++ )
    {
        nrVector3& n = m_Normals[ i ];
        nrColor& c = m_Colors[ i ];
        
        n = ( n + nrVector3( 1.0f, 1.0f, 1.0f ) ) * ( 1.0f / 2.0f );
        
        c = nrColor( n.x, n.y, n.z );
        
        if ( current % 100 == 0 )
        {
            Working( ( float )current / ( float )total );
        }
        
        current++;
    }
    
    // Re-index the colors.
    for( i = 0; i < m_NumTriangles; i++ )
    {
        nrmTriangle& t = m_Triangles[ i ];
        
        t.m_Vertices[ 0 ].m_ColorIndex = t.m_Vertices[ 0 ].m_VectorIndex;
        t.m_Vertices[ 1 ].m_ColorIndex = t.m_Vertices[ 1 ].m_VectorIndex;
        t.m_Vertices[ 2 ].m_ColorIndex = t.m_Vertices[ 2 ].m_VectorIndex;
        
        if ( current % 100 == 0 )
        {
            Working( ( float )current / ( float )total );
        }
        
        current++;
    }
    
    Working( 1.0f );
    printf( "\nFinished.\n" );
}

////////////////////////////////////////////////////////////////////////////

void nrModel::SetColor( int n, nrColor &c ) 
{
    assert( n >= 0 );
    assert( n < m_NumColors );
    
    m_Colors[ n ] = c;
}

////////////////////////////////////////////////////////////////////////////

void nrModel::Working( float percent )
{
    printf( "\rWorking...%3.0f%%.", percent * 100 );
}

////////////////////////////////////////////////////////////////////////////

void nrModel::SetName( const char* name )
{
    assert( name );
    
    delete [] m_Name;
    m_Name = new char[ strlen( name ) + 1 ];
    if ( m_Name == NULL )
    {
        printf( "nrModel::SetName(): Out of memory.\n" );
    }
    else
    {
        strcpy( m_Name, name );
    }
}

////////////////////////////////////////////////////////////////////////////
