////////////////////////////////////////////////////////////////////////////
//
// nrModel.h
//
// Nate Robins, 2001.
//
// Model class.
//
////////////////////////////////////////////////////////////////////////////

#ifndef NRMODEL_H
#define NRMODEL_H


////////////////////////////////////////////////////////////////////////////
// Includes
////////////////////////////////////////////////////////////////////////////

#include "nrColor.h"


////////////////////////////////////////////////////////////////////////////
// Classes
////////////////////////////////////////////////////////////////////////////

class nrVector2;
class nrVector3;
class nrModel;

////////////////////////////////////////////////////////////////////////////

class nrmVertex
{
public:
    int m_VectorIndex;      // Index into nrModel vectors list.
    int m_NormalIndex;      // Index into nrModel normals list.
    int m_ColorIndex;       // Index into nrModel colors list.
    int m_TexCoordIndex;    // Index into nrModel texcoords list.
};

////////////////////////////////////////////////////////////////////////////

class nrmTriangle
{
public:
    nrmTriangle();
    ~nrmTriangle();
    
    // Return the n'th (0,1,2) vector/etc of the triangle.
    const nrVector3&   Vector( int n ) const;
    const nrVector3&   Normal( int n ) const;
    const nrColor&     Color( int n ) const;
    const nrVector2&   TexCoord( int n ) const;
    const nrVector3&   FacetNormal( void ) const;
    
public:
    nrModel*  m_Model;               // The model this triangle is in.
    nrmVertex m_Vertices[ 3 ];       // Array of vertices for the triangle.
    int       m_FacetNormalIndex;    // Index into nrModel facetnormals list.
};

////////////////////////////////////////////////////////////////////////////

class nrmMaterial
{
public:
    nrmMaterial();
    ~nrmMaterial();
    
public:
    // Return the name of the group.
    const char* Name( void ) const;

private:
    // Set the materials name.
    void SetName( const char* name );
    
public:
    char*   m_Name;         // Name of the material
    nrColor m_Diffuse;      // Diffuse color
    nrColor m_Ambient;      // Ambient color
    nrColor m_Specular;     // Specular color
    nrColor m_Emissive;     // Emissive color
    float   m_Shininess;    // Specular exponent.
};

////////////////////////////////////////////////////////////////////////////

class nrmGroup
{
    friend class nrModelLoaderOBJ;

public:
    nrmGroup();
    ~nrmGroup();
    
public:
    // Return the name of the group.
    const char* Name( void ) const;
    
    // Return the number of triangles in the group.
    int NumTriangles( void ) const;
    
    // Return the n'th triangle in the group.
    const nrmTriangle& Triangle( int n ) const;
    
    // Return the groups material.
    const nrmMaterial& Material( void ) const;
    
private:
    // Set the name of the group.
    void SetName( const char* name );

    // Set the material index of the group.
    void SetMaterial( int index );
    
private:
    char*       m_Name;                 // Name of the group.
    nrModel*    m_Model;                // The model this group is in.
    int         m_NumTriangleIndices;   // Number of triangles in the group.
    int*        m_TriangleIndices;      // Indices into nrModel triangles list.
    int         m_MaterialIndex;        // Index into nrModel materials list.
};

////////////////////////////////////////////////////////////////////////////

class nrModel
{
    friend class nrModelLoader;
    friend class nrModelLoaderOBJ;
    friend class nrModelLoaderMTL;
    
    friend class nrmTriangle;
    friend class nrmGroup;
    
public:
    enum
    {
        RENDER_COLORS       = ( 1 << 0 ),
        RENDER_FACETNORMALS = ( 1 << 1 ),
        RENDER_NORMALS      = ( 1 << 2 ),
        RENDER_TEXCOORDS    = ( 1 << 3 ),
    };
    
public:
    nrModel();
    ~nrModel();
    
public:
    // Returns the number of vectors/etc in the model.
    int NumVectors( void ) const;
    int NumNormals( void ) const;
    int NumColors( void ) const;
    int NumTexCoords( void ) const;
    int NumFacetNormals( void ) const;
    int NumTriangles( void ) const;
    int NumMaterials( void ) const;
    int NumGroups( void ) const;
    
    // Returns the n'th vector/etc in the model.
    const nrVector3&   Vector( int n ) const;
    const nrVector3&   Normal( int n ) const;
    const nrColor&     Color( int n ) const;
    const nrVector2&   TexCoord( int n ) const;
    const nrVector3&   FacetNormal( int n ) const;
    const nrmTriangle& Triangle( int n ) const;
    const nrmMaterial& Material( int n ) const;
    const nrmGroup&    Group( int n ) const;
    
    void SetColor( int n, nrColor &c );
    
    // Returns the dimensions of the model.
    void Dimensions( nrVector3& mins, nrVector3& maxs ) const;
    
public:
    // Creates a facet normal for each triangle in the model.
    void CreateFacetNormals( void );
    
    // Creates smooth normals across the model.
    void CreateNormals( void );
    
    // Creates colors (from normals) across the model.
    void CreateColors( void );
    
public:
    // Render the model.
    void Render( unsigned int flags = 0 ) const;
    
    // Render a single group.
    void RenderGroup( int n, unsigned int flags = 0 ) const;
    
private:
    // Set the models name.
    void SetName( const char* name );
    
    // Display working progress.
    void Working( float percent );
    
private:
    char*       m_Name;            // Name of the model.
    char*       m_Path;            // Path of the model file.
    
    int         m_NumVectors;      // Number of vectors in the model.
    int         m_NumNormals;      // Number of normals in the model.
    int         m_NumColors;       // Number of colors in the model.
    int         m_NumTexCoords;    // Number of texture coordinate in the model.
    int         m_NumFacetNormals; // Number of facet normals in the model.
    int         m_NumTriangles;    // Number of triangles in the model.
    int         m_NumMaterials;    // Number of materials in the model.
    int         m_NumGroups;       // Number of groups in the model.
    
    nrVector3*   m_Vectors;        // Array of vectors (vertex positions).
    nrVector3*   m_Normals;        // Array of normals.
    nrColor*     m_Colors;         // Array of colors.
    nrVector2*   m_TexCoords;      // Array of texture coordinates.
    nrVector3*   m_FacetNormals;   // Array of facet normals.
    nrmTriangle* m_Triangles;      // Array of triangles.
    nrmMaterial* m_Materials;      // Array of materials.
    nrmGroup*    m_Groups;         // Array of of groups.
};

////////////////////////////////////////////////////////////////////////////

#endif        // NRMODEL_H
