////////////////////////////////////////////////////////////////////////////
//
// nrModelLoaderMTL.cpp
//
// Nate Robins, January 2002.
//
// Wavefront material file format reader.
//
////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////
// Includes
////////////////////////////////////////////////////////////////////////////

#include "nrModelLoaderMTL.h"

#include "nrModel.h"

#include <assert.h>
#include <string.h>
#include <stdlib.h>


////////////////////////////////////////////////////////////////////////////
// Defines
////////////////////////////////////////////////////////////////////////////

#define LINE_SIZE        1024


////////////////////////////////////////////////////////////////////////////
// Implementation
////////////////////////////////////////////////////////////////////////////

nrModelLoaderMTL::nrModelLoaderMTL()
{
    m_Line = new char[ LINE_SIZE ];
    assert( m_Line );
}

////////////////////////////////////////////////////////////////////////////

nrModelLoaderMTL::~nrModelLoaderMTL()
{
    delete [] m_Line;
}

////////////////////////////////////////////////////////////////////////////

#define NEW( ptr, type, num ) \
    ptr = new type[ num ]; \
    \
    if ( ptr == NULL ) \
    { \
        fprintf( stderr, "nrModelLoaderMTL::Read(): Out of memory.\n" ); \
        return NULL; \
    }

////////////////////////////////////////////////////////////////////////////

bool nrModelLoaderMTL::Read( const char* material_file, nrModel* model )
{
    nrModelLoaderMTL l;
    
    char* file_name = new char[ strlen( model->m_Path ) + strlen( material_file ) + 1 ];
    strcpy( file_name, model->m_Path );
    strcat( file_name, material_file );
    
    l.m_File = fopen( file_name, "rb" );
    if ( l.m_File == NULL )
    {
        fprintf( stderr, "nrModelLoaderMTL::Read(): Unable to open material file \"%s\".\n", file_name );
        delete file_name;
        return false;
    }
    
    delete file_name;
    
    l.m_Model = model;
    
    rewind( l.m_File );
    fseek( l.m_File, 0, SEEK_END );
    l.m_TotalBytes = ftell( l.m_File ) * 2;        // 2 passes.
    l.m_NumBytes = 0;
    
    l.FirstPass();
    
    // Allocate memory for vertices/etc.
    if ( l.m_Model->m_NumMaterials > 0 )
    {
        NEW( l.m_Model->m_Materials, nrmMaterial, l.m_Model->m_NumMaterials );
    }
    
    l.SecondPass();
    
    fclose( l.m_File );

    return true;
}

////////////////////////////////////////////////////////////////////////////

#undef NEW

////////////////////////////////////////////////////////////////////////////

void nrModelLoaderMTL::Loading( void )
{
    int percent = m_NumBytes * 100 / m_TotalBytes;
//    fprintf( stderr, "% 4d%%\r", percent );
}

////////////////////////////////////////////////////////////////////////////

void nrModelLoaderMTL::Rewind( void )
{
    m_LineNum = 0;
    
    rewind( m_File );
}

////////////////////////////////////////////////////////////////////////////

bool nrModelLoaderMTL::NextLine( void )
{
    if ( fgets( m_Line, LINE_SIZE, m_File ) == NULL )
    {
        Loading();
        
        return false;
    }
    else
    {
        m_NumBytes += strlen( m_Line );
        m_LineNum++;
        
        if ( m_LineNum % 1000 == 0 )
        {
            Loading();
        }
        
        return true;
    }
}

////////////////////////////////////////////////////////////////////////////

void nrModelLoaderMTL::FirstPass( void )
{
    int num_materials = 1;  // always at least one - default.
    
    Rewind();
    
    while ( NextLine() )
    {
        char* directive = strtok( m_Line, " \t\r\n" );
        char* line      = strtok( NULL, "\r\n" );
        
        if ( directive == NULL || line == NULL )
        {
            continue;
        }
        
        switch ( directive[ 0 ] )
        {
        case 'n':        // newmtl
            num_materials++;
            break;
            
        case 'K': {        // vertex - can be one of v, vn, vt
            switch ( directive[ 1 ] )
            {
            case 'a':        {        // Ka = ambient
                float r, g, b;
                if ( sscanf( line, "%f %f %f\n", &r, &g, &b ) != 3 )
                {
                    fprintf( stderr, "nrModelLoaderOBJ::Read(): Line %d - Malformed %s directive.\n", m_LineNum, directive );
                }
                              }
                break;
                
            case 'd':        {        // Kd = diffuse
                float r, g, b;
                if ( sscanf( line, "%f %f %f\n", &r, &g, &b ) != 3 )
                {
                    fprintf( stderr, "nrModelLoaderOBJ::Read(): Line %d - Malformed %s directive.\n", m_LineNum, directive );
                }
                             }
                break;
                
            case 's':        {        // Ks = specular
                float r, g, b;
                if ( sscanf( line, "%f %f %f\n", &r, &g, &b ) != 3 )
                {
                    fprintf( stderr, "nrModelLoaderOBJ::Read(): Line %d - Malformed %s directive.\n", m_LineNum, directive );
                }
                             }
                break;
                
            default:
                fprintf( stderr, "nrModelLoaderMTL::Read(): Line %d - Unknown color directive \"%s\" (ignored).\n", m_LineNum, directive );
                break;
            }
                  }
            break;
            
        case 'N':        // Ns = specular power (0-1000).
            float s;
            if ( sscanf( line, "%f\n", &s ) != 1 )
            {
                fprintf( stderr, "nrModelLoaderOBJ::Read(): Line %d - Malformed %s directive.\n", m_LineNum, directive );
            }
            break;

        case '#':       // # = comment.
        case 'i':       // illum = ?
        case 'd':       // d = ?
            break;
            
        default:
            fprintf( stderr, "nrModelLoaderMTL::Read(): Line %d - Unknown directive \"%s\" (ignored).\n", m_LineNum, directive );
            break;
        }
    }
    
    // Record all the information gathered.
    m_Model->m_NumMaterials = num_materials;
}

////////////////////////////////////////////////////////////////////////////

void nrModelLoaderMTL::SecondPass( void )
{
    Rewind();
    
    int num_materials = 1;
    nrmMaterial* material = &( m_Model->m_Materials[ 0 ] );
    
    // Fill out the default material.
    material->m_Diffuse = nrColor( 0.8f, 0.8f, 0.8f );
    material->m_Ambient = nrColor( 0.2f, 0.2f, 0.2f );
    material->m_Specular = nrColor( 0, 0, 0 );
    material->m_Emissive = nrColor( 0, 0, 0 );
    material->m_Shininess = 0;
    
    while ( NextLine() )
    {
        char* directive = strtok( m_Line, " \t\r\n" );
        char* line      = strtok( NULL, "\r\n" );
        
        if ( directive == NULL || line == NULL )
        {
            continue;
        }
        
        switch ( directive[ 0 ] )
        {
        case 'n': {        // newmtl
            material = &( m_Model->m_Materials[ num_materials ] );
            num_materials++;
            
            material->m_Name = new char[ strlen( line ) + 1 ];
            strcpy( material->m_Name, line );
                  }
            break;
            
        case 'K': {        // vertex - can be one of v, vn, vt
            switch ( directive[ 1 ] )
            {
            case 'a':        {        // Ka = ambient
                float r, g, b;
                sscanf( line, "%f %f %f\n", &r, &g, &b );
                material->m_Ambient = nrColor( r, g, b );
                              }
                break;
                
            case 'd':        {        // Kd = diffuse
                float r, g, b;
                sscanf( line, "%f %f %f\n", &r, &g, &b );
                material->m_Diffuse = nrColor( r, g, b );
                             }
                break;
                
            case 's':        {        // Ks = specular
                float r, g, b;
                sscanf( line, "%f %f %f\n", &r, &g, &b );
                material->m_Specular = nrColor( r, g, b );
                             }
                break;
                
            default:
                break;
            }
                  }
            break;
            
        case 'N': {        // Ns - specular power (shininess) (0-1000, so scale to 0-1).
            float s;
            sscanf( line, "%f\n", &s );
            s = s / 1000.0f;
            material->m_Shininess = s;
                  }
            break;
            
        default:
            break;
        }
    }
    
    // Sanity check.
    assert( num_materials == m_Model->m_NumMaterials );
}

////////////////////////////////////////////////////////////////////////////
