////////////////////////////////////////////////////////////////////////////
//
// nrModelLoaderOBJ.cpp
//
// Nate Robins, 2001.
//
// Wavefront OBJ model file format reader.
//
////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////
// Includes
////////////////////////////////////////////////////////////////////////////

#include "nrModelLoaderOBJ.h"
#include "nrModelLoaderMTL.h"

#include "nrModel.h"
#include "nrVector2.h"
#include "nrVector3.h"

#include <assert.h>
#include <ctype.h>
#include <math.h>
#include <string.h>


////////////////////////////////////////////////////////////////////////////
// Defines
////////////////////////////////////////////////////////////////////////////

#define LINE_SIZE        1024


////////////////////////////////////////////////////////////////////////////
// Implementation
////////////////////////////////////////////////////////////////////////////

nrModelLoaderOBJ::nrModelLoaderOBJ()
{
    m_Line = new char[ LINE_SIZE ];
    assert( m_Line );
}

////////////////////////////////////////////////////////////////////////////

nrModelLoaderOBJ::~nrModelLoaderOBJ()
{
    delete [] m_Line;
}

////////////////////////////////////////////////////////////////////////////

#define NEW( ptr, type, num ) \
    ptr = new type[ num ]; \
    \
    if ( ptr == NULL ) \
    { \
        fprintf( stderr, "nrModelLoaderOBJ::NewFromFile(): Out of memory.\n" ); \
        delete m_Model; \
        return NULL; \
    }

////////////////////////////////////////////////////////////////////////////

nrModel* nrModelLoaderOBJ::NewFromFile( const char* file_name )
{        
    FILE* file = fopen( file_name, "rb" );
    if ( file == NULL )
    {
        fprintf( stderr, "nrModelLoaderOBJ::NewFromFile(): Unable to open file \"%s\".\n", file_name );
        return NULL;
    }
    
    nrModel* m_Model = new nrModel;
    if ( m_Model == NULL )
    {
        return NULL;
    }
    else
    {
        nrModelLoaderOBJ l;
        
        l.m_File = file;
        
        rewind( l.m_File );
        fseek( l.m_File, 0, SEEK_END );
        l.m_TotalBytes = ftell( l.m_File ) * 3;        // 3 passes.
        l.m_NumBytes = 0;
        
        l.m_Model = m_Model;
        
        // Parse the path from the file name.
        {
            char drive[ _MAX_DRIVE ];
            char directory[ _MAX_DIR ];
            char extension[ _MAX_EXT ];
            char filename[ _MAX_FNAME ];
            
            _splitpath( file_name, drive, directory, filename, extension );
            
            l.m_Model->m_Path = new char[ strlen( directory ) + 1 ];
            strcpy( l.m_Model->m_Path, directory );
        }
        
        l.FirstPass();
        
        // Allocate memory for vertices/etc.
        if ( m_Model->m_NumVectors > 0 )
        {
            NEW( m_Model->m_Vectors, nrVector3, m_Model->m_NumVectors );
        }
        if ( m_Model->m_NumNormals > 0 )
        {
            NEW( m_Model->m_Normals, nrVector3, m_Model->m_NumNormals );
        }
        if ( m_Model->m_NumTexCoords > 0 )
        {
            NEW( m_Model->m_TexCoords, nrVector2, m_Model->m_NumTexCoords );
        }
        if ( m_Model->m_NumTriangles > 0 )
        {
            NEW( m_Model->m_Triangles, nrmTriangle, m_Model->m_NumTriangles );
        }
        if ( m_Model->m_NumGroups > 0 )
        {
            NEW( m_Model->m_Groups, nrmGroup, m_Model->m_NumGroups );
        }
        if ( m_Model->m_NumMaterials == 0 )
        {
            // Must have at least one (default) material.
            m_Model->m_NumMaterials = 1;
            NEW( m_Model->m_Materials, nrmMaterial, m_Model->m_NumMaterials );
            m_Model->m_Materials[ 0 ].m_Ambient = nrColor( 0, 0, 0 );
            m_Model->m_Materials[ 0 ].m_Ambient = nrColor( 0, 0, 0 );
            m_Model->m_Materials[ 0 ].m_Ambient = nrColor( 0, 0, 0 );
            m_Model->m_Materials[ 0 ].m_Ambient = nrColor( 0, 0, 0 );
        }
        
        l.SecondPass();
        
        // Combine groups of the same name into a single group.
        int num_groups = m_Model->m_NumGroups;
        int i;
        for (  i = 0; i < m_Model->m_NumGroups; i++ )
        {
            nrmGroup* group = &( m_Model->m_Groups[ i ] );
            
            if ( group->m_Name == NULL )
            {
                continue;
            }
            
            for ( int j = i + 1; j < m_Model->m_NumGroups; j++ )
            {
                nrmGroup* match = &( m_Model->m_Groups[ j ] );
                if ( match->m_Name && strcmp( group->m_Name, match->m_Name ) == 0 )
                {
                    group->m_NumTriangleIndices += match->m_NumTriangleIndices;
                    
                    delete [] match->m_Name;
                    match->m_Name = NULL;
                    
                    num_groups--;
                }
            }
        }
        
        // Compress the groups list.
        nrmGroup* groups = NULL;
        if ( num_groups > 0 )
        {
            NEW( groups, nrmGroup, num_groups );
        }
        
        int n = 0;
        for ( i = 0; i < m_Model->m_NumGroups; i++ )
        {
            nrmGroup* src = &( m_Model->m_Groups[ i ] );
            
            if ( src->m_Name != NULL )
            {
                assert( n < num_groups );
                
                nrmGroup* dst = &( groups[ n ] );
                
                dst->SetName( src->m_Name );
                dst->m_NumTriangleIndices = src->m_NumTriangleIndices;
                dst->m_MaterialIndex = src->m_MaterialIndex;
                
                n++;
            }
        }
        
        delete [] m_Model->m_Groups;
        m_Model->m_NumGroups = num_groups;
        m_Model->m_Groups = groups;
        
        // Allocate memory for the triangle indices in each group & set the
        // model pointer in the group.
        for ( i = 0; i < m_Model->m_NumGroups; i++ )
        {
            nrmGroup* group = &( m_Model->m_Groups[ i ] );
            NEW( group->m_TriangleIndices, int, group->m_NumTriangleIndices );
            group->m_Model = m_Model;
        }
        
        l.ThirdPass();
    
        fclose( file );

        return m_Model;
    }
}

////////////////////////////////////////////////////////////////////////////

#undef NEW

////////////////////////////////////////////////////////////////////////////

void nrModelLoaderOBJ::Loading( void )
{
    int percent = m_NumBytes * 100 / m_TotalBytes;
    fprintf( stderr, "% 4d%%\r", percent );
}

////////////////////////////////////////////////////////////////////////////

void nrModelLoaderOBJ::Rewind( void )
{
    m_LineNum = 0;
    
    rewind( m_File );
}

////////////////////////////////////////////////////////////////////////////

bool nrModelLoaderOBJ::NextLine( void )
{
    if ( fgets( m_Line, LINE_SIZE, m_File ) == NULL )
    {
        Loading();
        
        return false;
    }
    else
    {
        m_NumBytes += strlen( m_Line );
        m_LineNum++;

        if ( m_LineNum % 1000 == 0 )
        {
            Loading();
        }

        return true;
    }
}

////////////////////////////////////////////////////////////////////////////

void nrModelLoaderOBJ::FirstPass( void )
{
    int num_vectors   = 0;
    int num_normals   = 0;
    int num_texcoords = 0;
    int num_triangles = 0;
    int num_groups    = 1;        // There is always at least one "default" group.
    
    Rewind();
    
    while ( NextLine() )
    {
        char* directive = strtok( m_Line, " \t\r\n" );
        char* line      = strtok( NULL, "\r\n" );
        int   c;
        
        if ( directive == NULL || line == NULL )
        {
            continue;
        }
        
        switch ( directive[ 0 ] )
        {
        case 'v': {        // vertex - can be one of v, vn, vt
            switch ( directive[ 1 ] )
            {
            case '\0':        // v = vertex
                num_vectors++;
                break;
                
            case 'n':        // vn = vertex normal
                num_normals++;
                break;
                
            case 't':        // vt = vertex texcoord
                num_texcoords++;
                break;
                
            default:
                fprintf( stderr, "nrModelLoaderOBJ::FirstPass(): Line %d - Unknown vertex directive (ignored).\n", m_LineNum );
                break;
            }
                  }
            break;
            
        case 'f': {        // f = face - can be one of %d, %d//%d, %d/%d, %d/%d/%d
            int v = 0;
            int t = 0;
            int n = 0;
            if ( strstr( line, "//" ) )
            {
                // v//n        (vertex/normal)
                int r = 0;
                r += sscanf( line, "%d//%d%n", &v, &n, &c );        line += c;
                r += sscanf( line, "%d//%d%n", &v, &n, &c );        line += c;
                r += sscanf( line, "%d//%d%n", &v, &n, &c );        line += c;
                if ( r == 6 )
                {
                    num_triangles++;
                    while ( sscanf( line, "%d//%d%n", &v, &n, &c ) == 2 )
                    {
                        line += c;
                        num_triangles++;
                    }
                }
                else
                {
                    fprintf( stderr, "nrModelLoaderOBJ::NewFromFile(): Line %d - Malformed face directive.\n", m_LineNum );
                }
            }
            else if ( sscanf( line, "%d/%d/%d%n", &v, &t, &n, &c ) == 3 )
            {
                line += c;
                
                // v/t/n (vertex/texcoord/normal)
                int r = 3;
                r += sscanf( line, "%d/%d/%d%n", &v, &t, &n, &c );        line += c;
                r += sscanf( line, "%d/%d/%d%n", &v, &t, &n, &c );        line += c;
                if ( r == 9 )
                {
                    num_triangles++;
                    while ( sscanf( line, "%d/%d/%d%n", &v, &t, &n, &c ) == 3 ) 
                    {
                        line += c;
                        num_triangles++;
                    }
                }
                else
                {
                    fprintf( stderr, "nrModelLoaderOBJ::NewFromFile(): Line %d - Malformed face directive.\n", m_LineNum );
                }
            }
            else if ( sscanf( line, "%d/%d%n", &v, &t, &c ) == 2 )
            {
                line += c;
                
                // v/t (vertex/texcoord)
                int r = 2;
                r += sscanf( line, "%d/%d%n", &v, &t, &c );        line += c;
                r += sscanf( line, "%d/%d%n", &v, &t, &c );        line += c;
                if ( r == 6 )
                {
                    num_triangles++;
                    while ( sscanf( line, "%d/%d%n", &v, &t, &c ) == 2 )
                    {
                        line += c;
                        num_triangles++;
                    }
                }
                else
                {
                    fprintf( stderr, "nrModelLoaderOBJ::NewFromFile(): Line %d - Malformed face directive.\n", m_LineNum );
                }
            }
            else if ( sscanf( line, "%d%n", &v, &c ) == 1 )
            {
                line += c;
                
                // v (vertex)
                int r = 0;
                r += sscanf( line, "%d%n", &v, &c );        line += c;
                r += sscanf( line, "%d%n", &v, &c );        line += c;
                if ( r == 2 )
                {
                    num_triangles++;
                    while ( sscanf( line, "%d%n", &v, &c ) == 1 )
                    {
                        line += c;
                        num_triangles++;
                    }
                }
                else
                {
                    fprintf( stderr, "nrModelLoaderOBJ::NewFromFile(): Line %d - Malformed face directive.\n", m_LineNum );
                }
            }
            else
            {
                fprintf( stderr, "nrModelLoaderOBJ::NewFromFile(): Line %d - Malformed face directive.\n", m_LineNum );
            }
                  }
            break;
            
        case 'g':        // g = group
            num_groups++;
            break;
            
        case 'm':   {    // m = mtllib (material library)
            nrModelLoaderMTL::Read( line, m_Model );
                    }
            break;
            
        case '#':        // # = comment
        case 'u':        // u = usemtl (use material)
        case 's':        // s = smoothing group
            break;
            
        default:
            fprintf( stderr, "nrModelLoaderOBJ::NewFromFile(): Line %d - Unknown directive \"%s\" (ignored).\n", m_LineNum, directive );
            break;
        }
    }
    
    // Record all the information gathered.
    m_Model->m_NumVectors   = num_vectors;
    m_Model->m_NumNormals   = num_normals;
    m_Model->m_NumTexCoords = num_texcoords;
    m_Model->m_NumTriangles = num_triangles;
    m_Model->m_NumGroups    = num_groups;
}

////////////////////////////////////////////////////////////////////////////

void nrModelLoaderOBJ::SecondPass( void )
{
    Rewind();
    
    int num_groups = 0;
    int c;
    
    nrmGroup* group = &( m_Model->m_Groups[ num_groups ] );
    group->SetName( "default" );
    num_groups++;
    
    while ( NextLine() )
    {
        char* directive = strtok( m_Line, " \t\r\n" );
        char* line      = strtok( NULL, "\r\n" );
        
        if ( directive == NULL || line == NULL )
        {
            continue;
        }
        
        switch ( directive[ 0 ] )
        {
#if 0        // This should work, but the stuff in the else is safer.
        case 'f': // f = face
            while ( strtok( line, " \t\r\n" ) )
            {
                line = NULL;
                group->m_NumTriangleIndices++;
            }
            group->m_NumTriangleIndices -= 2;
            break;
#else
        case 'f': {        // f = face - can be one of %d, %d//%d, %d/%d, %d/%d/%d
            int v0, v1, v2;        v0 = v1 = v2 = 0;
            int t0, t1, t2; t0 = t1 = t2 = 0;
            int n0, n1, n2; n0 = n1 = n2 = 0;
            if ( strstr( line, "//" ) )
            {
                // v//n        (vertex/normal)
                int r = 0;
                r += sscanf( line, "%d//%d%n", &v0, &n0, &c );        line += c;
                r += sscanf( line, "%d//%d%n", &v1, &n1, &c );        line += c;
                r += sscanf( line, "%d//%d%n", &v2, &n2, &c );        line += c;
                if ( r == 6 )
                {
                    group->m_NumTriangleIndices++;
                    
                    while ( sscanf( line, "%d//%d%n", &v2, &n2, &c ) == 2 )
                    {
                        line += c;
                        group->m_NumTriangleIndices++;
                    }
                }
            }
            else if ( sscanf( line, "%d/%d/%d%n", &v0, &t0, &n0, &c ) == 3 )
            {
                line += c;
                
                // v/t/n (vertex/texcoord/normal)
                int r = 3;
                r += sscanf( line, "%d/%d/%d%n", &v1, &t1, &n1, &c );        line += c;
                r += sscanf( line, "%d/%d/%d%n", &v2, &t2, &n2, &c );        line += c;
                if ( r == 9 )
                {
                    group->m_NumTriangleIndices++;
                    
                    while ( sscanf( line, "%d/%d/%d%n", &v2, &t2, &n2, &c ) == 3 ) 
                    {
                        line += c;
                        group->m_NumTriangleIndices++;
                    }
                }
            }
            else if ( sscanf( line, "%d/%d%n", &v0, &t0, &c ) == 2 )
            {
                line += c;
                
                // v/t (vertex/texcoord)
                int r = 2;
                r += sscanf( line, "%d/%d%n", &v1, &t1, &c );        line += c;
                r += sscanf( line, "%d/%d%n", &v2, &t2, &c );        line += c;
                if ( r == 6 )
                {
                    group->m_NumTriangleIndices++;
                    
                    while ( sscanf( line, "%d/%d%n", &v2, &t2, &c ) == 2 )
                    {
                        line += c;
                        group->m_NumTriangleIndices++;
                    }
                }
            }
            else if ( sscanf( line, "%d%n", &v0, &c ) == 1 )
            {
                line += c;
                
                // v (vertex)
                int r = 0;
                r += sscanf( line, "%d%n", &v1, &c );        line += c;
                r += sscanf( line, "%d%n", &v2, &c );        line += c;
                if ( r == 2 )
                {
                    group->m_NumTriangleIndices++;
                    
                    while ( sscanf( line, "%d%n", &v2, &c ) == 1 )
                    {
                        line += c;
                        group->m_NumTriangleIndices++;
                    }
                }
            }
                  }
            break;
#endif
            
        case 'g': // g = group
            group = &( m_Model->m_Groups[ num_groups ] );
            group->SetName( line );
            num_groups++;
            break;
            
        default:
            break;
        }
    }
    
    // Sanity check.
    assert( num_groups == m_Model->m_NumGroups );
}

////////////////////////////////////////////////////////////////////////////

nrmGroup* nrModelLoaderOBJ::FindGroup( const char* name )
{
    for ( int i = 0; i < m_Model->m_NumGroups; i++ )
    {
        nrmGroup* group = &( m_Model->m_Groups[ i ] );
        
        if ( strcmp( group->m_Name, name ) == 0 )
        {
            return group;
        }
    }
    
    // Sanity check.
    assert( false );
    return NULL;
}

////////////////////////////////////////////////////////////////////////////

int nrModelLoaderOBJ::FindMaterial( const char* name )
{
    assert( name );

    for ( int i = 0; i < m_Model->m_NumMaterials; i++ )
    {
        nrmMaterial* material = &( m_Model->m_Materials[ i ] );
        
        if ( material->m_Name && strcmp( material->m_Name, name ) == 0 )
        {
            return i;
        }
    }
    
    // Didn't find one - send back the default.
    fprintf( stderr, "nrModelLoaderOBJ::NewFromFile(): Line %d - Unknown material \"%s\" (using default).\n", m_LineNum, name );
    
    return 0;
}

////////////////////////////////////////////////////////////////////////////

nrmTriangle* nrModelLoaderOBJ::SetTriangle( int triangle, int v0, int n0, int t0, int v1, int n1, int t1, int v2, int n2, int t2 )
{
    nrmTriangle* t = &( m_Model->m_Triangles[ triangle ] );
    
    t->m_Vertices[ 0 ].m_VectorIndex   = v0 - 1;
    t->m_Vertices[ 0 ].m_NormalIndex   = n0 - 1;
    t->m_Vertices[ 0 ].m_TexCoordIndex = t0 - 1;
    
    t->m_Vertices[ 1 ].m_VectorIndex   = v1 - 1;
    t->m_Vertices[ 1 ].m_NormalIndex   = n1 - 1;
    t->m_Vertices[ 1 ].m_TexCoordIndex = t1 - 1;
    
    t->m_Vertices[ 2 ].m_VectorIndex   = v2 - 1;
    t->m_Vertices[ 2 ].m_NormalIndex   = n2 - 1;
    t->m_Vertices[ 2 ].m_TexCoordIndex = t2 - 1;
    
    t->m_Model = m_Model;
    
    return t;
}

////////////////////////////////////////////////////////////////////////////

nrmTriangle* nrModelLoaderOBJ::SetNextTriangle( nrmTriangle* p, int triangle, int v2, int n2, int t2 )
{
    int v0 = p->m_Vertices[ 0 ].m_VectorIndex;
    int n0 = p->m_Vertices[ 0 ].m_NormalIndex;
    int t0 = p->m_Vertices[ 0 ].m_TexCoordIndex;
    
    int v1 = p->m_Vertices[ 2 ].m_VectorIndex;
    int n1 = p->m_Vertices[ 2 ].m_NormalIndex;
    int t1 = p->m_Vertices[ 2 ].m_TexCoordIndex;
    
    nrmTriangle* t = &( m_Model->m_Triangles[ triangle ] );
    
    t->m_Vertices[ 0 ].m_VectorIndex   = v0;
    t->m_Vertices[ 0 ].m_NormalIndex   = n0;
    t->m_Vertices[ 0 ].m_TexCoordIndex = t0;
    
    t->m_Vertices[ 1 ].m_VectorIndex   = v1;
    t->m_Vertices[ 1 ].m_NormalIndex   = n1;
    t->m_Vertices[ 1 ].m_TexCoordIndex = t1;
    
    t->m_Vertices[ 2 ].m_VectorIndex   = v2 - 1;
    t->m_Vertices[ 2 ].m_NormalIndex   = n2 - 1;
    t->m_Vertices[ 2 ].m_TexCoordIndex = t2 - 1;
    
    t->m_Model = m_Model;
    
    return t;
}

////////////////////////////////////////////////////////////////////////////

void nrModelLoaderOBJ::AddTriangleToGroup( nrmGroup* group, int triangle )
{
    group->m_TriangleIndices[ group->m_NumTriangleIndices ] = triangle;
    group->m_NumTriangleIndices++;
}

////////////////////////////////////////////////////////////////////////////

void nrModelLoaderOBJ::ThirdPass( void )
{
    Rewind();
    
    int num_vectors   = 0;
    int num_normals   = 0;
    int num_texcoords = 0;
    int num_triangles = 0;
    
    // Clear all group triangle index counts, because they must be recounted
    // as they are filled in.
    nrmGroup* group;
    for ( int i = 0; i < m_Model->m_NumGroups; i++ )
    {
        group = &( m_Model->m_Groups[ i ] );
        group->m_NumTriangleIndices = 0;
    }
    
    // Set up the default group.
    group = &( m_Model->m_Groups[ 0 ] );

    // Set the default material.
    int material = 0;
    
    while ( NextLine() )
    {
        char* directive = strtok( m_Line, " \t\r\n" );
        char* line      = strtok( NULL, "\r\n" );
        int   c;
        
        if ( directive == NULL || line == NULL )
        {
            continue;
        }
        
        switch ( directive[ 0 ] )
        {
        case 'v': {        // vertex - can be one of v, vn, vt
            switch ( directive[ 1 ] )
            {
            case '\0':        {        // v = vertex
                nrVector3& v = m_Model->m_Vectors[ num_vectors ];
                sscanf( line, "%f %f %f\n", &v.x, &v.y, &v.z );
                num_vectors++;
                              }
                break;
                
            case 'n':        {        // vn = vertex normal
                nrVector3& n = m_Model->m_Normals[ num_normals ];
                sscanf( line, "%f %f %f\n", &n.x, &n.y, &n.z );
                num_normals++;
                             }
                break;
                
            case 't':        {        // vt = vertex texcoord
                nrVector2& t = m_Model->m_TexCoords[ num_texcoords ];
                sscanf( line, "%f %f\n", &t.x, &t.y );
                num_texcoords++;
                             }
                break;
                
            default:
                break;
            }
                  }
            break;
            
        case 'f': {        // f = face - can be one of %d, %d//%d, %d/%d, %d/%d/%d
            int v0, v1, v2;        v0 = v1 = v2 = 0;
            int t0, t1, t2; t0 = t1 = t2 = 0;
            int n0, n1, n2; n0 = n1 = n2 = 0;
            if ( strstr( line, "//" ) )
            {
                // v//n        (vertex/normal)
                int r = 0;
                r += sscanf( line, "%d//%d%n", &v0, &n0, &c );        line += c;
                r += sscanf( line, "%d//%d%n", &v1, &n1, &c );        line += c;
                r += sscanf( line, "%d//%d%n", &v2, &n2, &c );        line += c;
                if ( r == 6 )
                {
                    nrmTriangle* t = SetTriangle( num_triangles, v0, n0, t0, v1, n1, t1, v2, n2, t2 );
                    AddTriangleToGroup( group, num_triangles );
                    num_triangles++;
                    
                    while ( sscanf( line, "%d//%d%n", &v2, &n2, &c ) == 2 )
                    {
                        line += c;
                        t = SetNextTriangle( t, num_triangles, v2, n2, t2 );
                        AddTriangleToGroup( group, num_triangles );
                        num_triangles++;
                    }
                }
            }
            else if ( sscanf( line, "%d/%d/%d%n", &v0, &t0, &n0, &c ) == 3 )
            {
                line += c;
                
                // v/t/n (vertex/texcoord/normal)
                int r = 3;
                r += sscanf( line, "%d/%d/%d%n", &v1, &t1, &n1, &c );        line += c;
                r += sscanf( line, "%d/%d/%d%n", &v2, &t2, &n2, &c );        line += c;
                if ( r == 9 )
                {
                    nrmTriangle* t = SetTriangle( num_triangles, v0, n0, t0, v1, n1, t1, v2, n2, t2 );
                    AddTriangleToGroup( group, num_triangles );
                    num_triangles++;
                    
                    while ( sscanf( line, "%d/%d/%d%n", &v2, &t2, &n2, &c ) == 3 ) 
                    {
                        line += c;
                        t = SetNextTriangle( t, num_triangles, v2, n2, t2 );
                        AddTriangleToGroup( group, num_triangles );
                        num_triangles++;
                    }
                }
            }
            else if ( sscanf( line, "%d/%d%n", &v0, &t0, &c ) == 2 )
            {
                line += c;
                
                // v/t (vertex/texcoord)
                int r = 2;
                r += sscanf( line, "%d/%d%n", &v1, &t1, &c );        line += c;
                r += sscanf( line, "%d/%d%n", &v2, &t2, &c );        line += c;
                if ( r == 6 )
                {
                    nrmTriangle* t = SetTriangle( num_triangles, v0, n0, t0, v1, n1, t1, v2, n2, t2 );
                    AddTriangleToGroup( group, num_triangles );
                    num_triangles++;
                    
                    while ( sscanf( line, "%d/%d%n", &v2, &t2, &c ) == 2 )
                    {
                        line += c;
                        t = SetNextTriangle( t, num_triangles, v2, n2, t2 );
                        AddTriangleToGroup( group, num_triangles );
                        num_triangles++;
                    }
                }
            }
            else if ( sscanf( line, "%d%n", &v0, &c ) == 1 )
            {
                line += c;
                
                // v (vertex)
                int r = 0;
                r += sscanf( line, "%d%n", &v1, &c );        line += c;
                r += sscanf( line, "%d%n", &v2, &c );        line += c;
                if ( r == 2 )
                {
                    nrmTriangle* t = SetTriangle( num_triangles, v0, n0, t0, v1, n1, t1, v2, n2, t2 );
                    AddTriangleToGroup( group, num_triangles );
                    num_triangles++;
                    
                    while ( sscanf( line, "%d%n", &v2, &c ) == 1 )
                    {
                        line += c;
                        t = SetNextTriangle( t, num_triangles, v2, n2, t2 );
                        AddTriangleToGroup( group, num_triangles );
                        num_triangles++;
                    }
                }
            }
                  }
            break;
            
        case 'u': // u = usemtl
            material = FindMaterial( line );
            group->SetMaterial( material );
            break;
            
        case 'g': // g = group
            group = FindGroup( line );
            break;
            
        default:
            break;
        }
    }
    
    // Sanity check.
    assert( num_vectors   == m_Model->m_NumVectors );
    assert( num_normals   == m_Model->m_NumNormals );
    assert( num_triangles == m_Model->m_NumTriangles );
}

////////////////////////////////////////////////////////////////////////////

