////////////////////////////////////////////////////////////////////////////
//
// nrSurfaceBVH.cpp
//
// A class for a surface bounding volume hierarchy.
//
// Nate Robins, January 2002.
//
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Includes
////////////////////////////////////////////////////////////////////////////

#include "nrSurfaceBVH.h"

#include "nrHit.h"
#include "nrInterval.h"
#include "nrRay.h"


////////////////////////////////////////////////////////////////////////////
// Public
////////////////////////////////////////////////////////////////////////////

nrSurfaceBVH::nrSurfaceBVH( const nrBound& bound )
{
    m_Left = 0;
    m_Right = 0;

    m_Bound = bound;
}

////////////////////////////////////////////////////////////////////////////

nrSurfaceBVH::~nrSurfaceBVH( void )
{
    delete m_Left;
    delete m_Right;
}

////////////////////////////////////////////////////////////////////////////

bool nrSurfaceBVH::Hit( const nrRay& ray, nrInterval& interval, nrHit& hit ) const
{
    if ( m_Bound.Hit( ray, interval ) )
    {
        assert( m_Left != 0 && m_Right != 0 );
        
        nrHit hit_left;
        nrHit hit_right;
        
        bool hit_something_left = m_Left->Hit( ray, interval, hit_left );
        bool hit_something_right = m_Right->Hit( ray, interval, hit_right );
        
        if ( hit_something_left && hit_something_right )
        {
            if ( hit_left.t < hit_right.t )
            {
                hit = hit_left;
                return true;
            }
            else
            {
                hit = hit_right;
                return true;
            }
        }
        else if ( hit_something_left )
        {
            hit = hit_left;
            return true;
        }
        else if ( hit_something_right )
        {
            hit = hit_right;
            return true;
        }
        else
        {
            return false;
        }
    }
    else
    {
        return false;
    }
}

////////////////////////////////////////////////////////////////////////////

nrVector3 nrSurfaceBVH::Normal( const nrVector3& point ) const
{
    return nrVector3( 0, 0, 0 );
}

////////////////////////////////////////////////////////////////////////////
// Static
////////////////////////////////////////////////////////////////////////////

int nrSurfaceBVH::SortInX( const void* _a, const void* _b )
{
    const nrSurface*& sa = *( nrSurface** )_a;
    const nrSurface*& sb = *( nrSurface** )_b;
    
    const nrBound& a = sa->Bound();
    const nrBound& b = sb->Bound();
    
    if ( a.m_Minimums.x > b.m_Minimums.x )
    {
        return 1;
    }
    else if ( b.m_Minimums.x > a.m_Minimums.x )
    {
        return -1;
    }
    else
    {
        return 0;
    }
}

////////////////////////////////////////////////////////////////////////////

int nrSurfaceBVH::SortInY( const void* _a, const void* _b )
{
    const nrSurface*& sa = *( nrSurface** )_a;
    const nrSurface*& sb = *( nrSurface** )_b;
    
    const nrBound& a = sa->Bound();
    const nrBound& b = sb->Bound();
    
    if ( a.m_Minimums.y > b.m_Minimums.y )
    {
        return 1;
    }
    else if ( b.m_Minimums.y > a.m_Minimums.y )
    {
        return -1;
    }
    else
    {
        return 0;
    }
}

////////////////////////////////////////////////////////////////////////////

int nrSurfaceBVH::SortInZ( const void* _a, const void* _b )
{
    const nrSurface*& sa = *( nrSurface** )_a;
    const nrSurface*& sb = *( nrSurface** )_b;
    
    const nrBound& a = sa->Bound();
    const nrBound& b = sb->Bound();
    
    if ( a.m_Minimums.z > b.m_Minimums.z )
    {
        return 1;
    }
    else if ( b.m_Minimums.z > a.m_Minimums.z )
    {
        return -1;
    }
    else
    {
        return 0;
    }
}

////////////////////////////////////////////////////////////////////////////

nrSurface* nrSurfaceBVH::CreateTree( nrArray< nrSurface* >& surfaces )
{
    assert( surfaces.Length() > 0 );
    
    // If there is only a single surface in the list, return it as a leaf.
    if ( surfaces.Length() == 1 )
    {
        return surfaces[ 0 ];
    }
    
    // Compute the bounding volume of the parent (must enclose all children).
    nrBound bound = surfaces[ 0 ]->Bound();
    for ( int i = 1; i < surfaces.Length(); i++ )
    {
        const nrBound& b = surfaces[ i ]->Bound();
        
        if ( b.m_Minimums.x < bound.m_Minimums.x )
        {
            bound.m_Minimums.x = b.m_Minimums.x;
        }
        if ( b.m_Minimums.y < bound.m_Minimums.y )
        {
            bound.m_Minimums.y = b.m_Minimums.y;
        }
        if ( b.m_Minimums.z < bound.m_Minimums.z )
        {
            bound.m_Minimums.z = b.m_Minimums.z;
        }
        
        if ( b.m_Maximums.x > bound.m_Maximums.x )
        {
            bound.m_Maximums.x = b.m_Maximums.x;
        }
        if ( b.m_Maximums.y > bound.m_Maximums.y )
        {
            bound.m_Maximums.y = b.m_Maximums.y;
        }
        if ( b.m_Maximums.z > bound.m_Maximums.z )
        {
            bound.m_Maximums.z = b.m_Maximums.z;
        }
    }
    
    nrSurfaceBVH* parent = new nrSurfaceBVH( bound );
    assert( parent );
    
    // Sort the array of surfaces along a particular axis.
    static int axis = -1;
    axis = ( axis + 1 ) % 3;
    if ( axis == 0 )
    {
        surfaces.Sort( SortInX );
    }
    else if ( axis == 1 )
    {
        surfaces.Sort( SortInY );
    }
    else
    {
        surfaces.Sort( SortInZ );
    }
    
    // Split the array in half.
    int num_left = surfaces.Length() / 2;
    int num_right = surfaces.Length() - num_left;
    
    nrArray< nrSurface* > left( num_left );
    nrArray< nrSurface* > right( num_right );
    
    for ( i = 0; i < num_left; i++ )
    {
        left.Add( surfaces[ i ] );
    }
    for ( i = 0; i < num_right; i++ )
    {
        right.Add( surfaces[ num_left + i ] );
    }
    
    // Create the children.
    parent->m_Left = CreateTree( left );
    parent->m_Right = CreateTree( right );
    
    return parent;
}

////////////////////////////////////////////////////////////////////////////
