////////////////////////////////////////////////////////////////////////////
//
// nrSurfaceBox.cpp
//
// A class for a box.
//
// Nate Robins, January 2002.
//
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Includes
////////////////////////////////////////////////////////////////////////////

#include "nrSurfaceBox.h"

#include "nrHit.h"
#include "nrInterval.h"
#include "nrLog.h"
#include "nrMath.h"
#include "nrParser.h"
#include "nrRay.h"

#include <assert.h>
#include <stdio.h>


////////////////////////////////////////////////////////////////////////////
// Public
////////////////////////////////////////////////////////////////////////////

nrSurfaceBox::nrSurfaceBox( const nrVector3& p0, const nrVector3& p1, const nrMaterial& material )
{
    nrVector3 mins;
    nrVector3 maxs;
    
    if ( p0.x < p1.x )
    {
        mins.x = p0.x;
        maxs.x = p1.x;
    }
    else
    {
        mins.x = p1.x;
        maxs.x = p0.x;
    }

    if ( p0.y < p1.y )
    {
        mins.y = p0.y;
        maxs.y = p1.y;
    }
    else
    {
        mins.y = p1.y;
        maxs.y = p0.y;
    }

    if ( p0.z < p1.z )
    {
        mins.z = p0.z;
        maxs.z = p1.z;
    }
    else
    {
        mins.z = p1.z;
        maxs.z = p0.z;
    }
    
    m_P0 = mins;
    m_P1 = maxs;
    
    m_Bound.m_Minimums = mins;
    m_Bound.m_Maximums = maxs;
    
    m_Material = material;
}

////////////////////////////////////////////////////////////////////////////

nrSurfaceBox::~nrSurfaceBox( void )
{
}

////////////////////////////////////////////////////////////////////////////

nrVector3 nrSurfaceBox::Normal( const nrVector3& point ) const
{
    float x = nrMath::Abs( point.x );
    float y = nrMath::Abs( point.y );
    float z = nrMath::Abs( point.z );
    
    if ( x > y && x > z )
    {
        nrVector3 n = nrVector3( point.x, 0, 0 );
        return n.Unit();
    }
    else if ( y > x && y > z )
    {
        nrVector3 n = nrVector3( 0, point.y, 0 );
        return n.Unit();
    }
    else
    {
        nrVector3 n = nrVector3( 0, 0, point.z );
        return n.Unit();
    }
}

////////////////////////////////////////////////////////////////////////////

bool nrSurfaceBox::Hit( const nrRay& ray, nrInterval& interval, nrHit& hit ) const
{
    const nrVector3& d = ray.d;
    const nrVector3& o = ray.o;
    
    nrInterval x;
    nrInterval y;
    nrInterval z;
    
    if ( d.x > 0 ) 
    {
        float a = ( m_P0.x - o.x ) / d.x;
        float b = ( m_P1.x - o.x ) / d.x;
        x = nrInterval( a, b );
    }
    else
    {
        float a = ( m_P0.x - o.x ) / d.x;
        float b = ( m_P1.x - o.x ) / d.x;
        x = nrInterval( b, a );
    }

    if ( d.y > 0 ) 
    {
        float a = ( m_P0.y - o.y ) / d.y;
        float b = ( m_P1.y - o.y ) / d.y;
        y = nrInterval( a, b );
    }
    else
    {
        float a = ( m_P0.y - o.y ) / d.y;
        float b = ( m_P1.y - o.y ) / d.y;
        y = nrInterval( b, a );
    }
    
    if ( d.z > 0 ) 
    {
        float a = ( m_P0.z - o.z ) / d.z;
        float b = ( m_P1.z - o.z ) / d.z;
        z = nrInterval( a, b );
    }
    else
    {
        float a = ( m_P0.z - o.z ) / d.z;
        float b = ( m_P1.z - o.z ) / d.z;
        z = nrInterval( b, a );
    }
    
    nrInterval overlap = ( x.Overlap( y ) ).Overlap( z );
    
    /*
    float t0, t1;
    
    if ( x.m_Minimum > y.m_Minimum )
    {
        t0 = x.m_Minimum;
    }
    else
    {
        t0 = y.m_Minimum;
    }
    if ( z.m_Minimum > t0 )
    {
        t0 = z.m_Minimum;
    }
    
    if ( x.m_Maximum < y.m_Maximum )
    {
        t1 = x.m_Maximum;
    }
    else
    {
        t1 = y.m_Maximum;
    }
    if ( z.m_Maximum < t1 )
    {
        t1 = z.m_Maximum;
    }
    */
    
    if ( ! overlap.Empty() )
    {
        float t = overlap.m_Minimum;
        if ( interval.Includes( t ) )
        {
            hit = nrHit( this, t );
            return true;
        }
    }
    
    return false;
}

////////////////////////////////////////////////////////////////////////////

// !!! add checks for ensuring that at least a min and a max were read.

nrSurfaceBox* nrSurfaceBox::Parse( nrParser& parser )
{
    nrVector3  p0, p1;
    nrMaterial material;
    
    if ( parser.NextToken() == nrParser::TOKEN_STRING )
    {
        const char* name = parser.ReadString();
    }
    
    parser.ReadToken( nrParser::TOKEN_AGGREGATE_BEGIN );
    
    while ( parser.NextToken() == nrParser::TOKEN_KEY )
    {
        const char* key = parser.ReadKey();
        
        if ( parser.KeyMatches( key, "p0" ) )
        {
            parser.ReadToken( nrParser::TOKEN_LIST_BEGIN );
            p0.x = parser.ReadFloat();
            p0.y = parser.ReadFloat();
            p0.z = parser.ReadFloat();
            parser.ReadToken( nrParser::TOKEN_LIST_END );
        }
        else if ( parser.KeyMatches( key, "p1" ) )
        {
            parser.ReadToken( nrParser::TOKEN_LIST_BEGIN );
            p1.x = parser.ReadFloat();
            p1.y = parser.ReadFloat();
            p1.z = parser.ReadFloat();
            parser.ReadToken( nrParser::TOKEN_LIST_END );
        }
        else if ( parser.KeyMatches( key, "color" ) )
        {
            parser.ReadToken( nrParser::TOKEN_LIST_BEGIN );
            material.m_Diffuse.r = parser.ReadFloat();
            material.m_Diffuse.g = parser.ReadFloat();
            material.m_Diffuse.b = parser.ReadFloat();
            parser.ReadToken( nrParser::TOKEN_LIST_END );
            material.m_Ambient.r = material.m_Diffuse.r;
            material.m_Ambient.g = material.m_Diffuse.g;
            material.m_Ambient.b = material.m_Diffuse.b;
        }
        else if ( parser.KeyMatches( key, "material" ) )
        {
            nrMaterial* m = nrMaterial::Parse( parser );
            material = *m;
            delete m;
        }
        else
        {
            parser.ParseError( "unknown key." );
        }
    }
    
    parser.ReadToken( nrParser::TOKEN_AGGREGATE_END );
    
    if ( parser.Error() )
    {
        return 0;
    }
    
    nrSurfaceBox* box = new nrSurfaceBox( p0, p1, material );
    
    return box;
}

////////////////////////////////////////////////////////////////////////////
