////////////////////////////////////////////////////////////////////////////
//
// nrSurfaceSphere.cpp
//
// A class for a sphere.
//
// Nate Robins, January 2002.
//
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Includes
////////////////////////////////////////////////////////////////////////////

#include "nrSurfaceSphere.h"

#include "nrHit.h"
#include "nrInterval.h"
#include "nrLog.h"
#include "nrMath.h"
#include "nrParser.h"
#include "nrRay.h"

#include <assert.h>
#include <stdio.h>


////////////////////////////////////////////////////////////////////////////
// Public
////////////////////////////////////////////////////////////////////////////

nrSurfaceSphere::nrSurfaceSphere( const nrVector3& center, float radius, const nrMaterial& material )
{
    m_Center   = center;
    m_Radius   = radius;
    
    m_Bound.m_Minimums = nrVector3( center.x - radius, center.y - radius, center.z - radius );
    m_Bound.m_Maximums = nrVector3( center.x + radius, center.y + radius, center.z + radius );

    m_Material = material;
}

////////////////////////////////////////////////////////////////////////////

nrSurfaceSphere::~nrSurfaceSphere( void )
{
}

////////////////////////////////////////////////////////////////////////////

nrVector3 nrSurfaceSphere::Normal( const nrVector3& point ) const
{
    return ( point - m_Center ).Unit();
}

////////////////////////////////////////////////////////////////////////////

static float epsilon = 0.0001f;

bool nrSurfaceSphere::Hit( const nrRay& ray, nrInterval& interval, nrHit& hit ) const
{
    const nrVector3& d = ray.d;
    const nrVector3& o = ray.o;
    const nrVector3 co = o - m_Center;
    
    float A = d.Dot( d );       assert( ! nrMath::Equal( A, 0.0f ) );
    float B = 2 * d.Dot( co );
    float C = co.Dot( co ) - m_Radius * m_Radius;
    float discriminant = B * B - 4 * A * C;
    
    if ( discriminant >= epsilon )
    {
        float s = nrMath::Sqrt( discriminant );
        
        float t = ( -B - s ) / ( 2 * A );
        if ( interval.Includes( t ) ) 
        {
            hit = nrHit( this, t );
            return true;
        }
        
        t = ( -B + s ) / ( 2 * A );
        if ( interval.Includes( t ) )
        {
            hit = nrHit( this, t );
            return true;
        }
    }
    
    return false;
}

////////////////////////////////////////////////////////////////////////////

// !!! add checks for ensuring that at least a center and a radius were read.

nrSurfaceSphere* nrSurfaceSphere::Parse( nrParser& parser )
{
    nrVector3  center;
    float      radius;
    nrMaterial material;
    
    if ( parser.NextToken() == nrParser::TOKEN_STRING )
    {
        const char* name = parser.ReadString();
    }
    
    parser.ReadToken( nrParser::TOKEN_AGGREGATE_BEGIN );
    
    while ( parser.NextToken() == nrParser::TOKEN_KEY )
    {
        const char* key = parser.ReadKey();
        
        if ( parser.KeyMatches( key, "center" ) )
        {
            parser.ReadToken( nrParser::TOKEN_LIST_BEGIN );
            center.x = parser.ReadFloat();
            center.y = parser.ReadFloat();
            center.z = parser.ReadFloat();
            parser.ReadToken( nrParser::TOKEN_LIST_END );
        }
        else if ( parser.KeyMatches( key, "radius" ) )
        {
            radius = parser.ReadFloat();
        }
        else if ( parser.KeyMatches( key, "color" ) )
        {
            parser.ReadToken( nrParser::TOKEN_LIST_BEGIN );
            material.m_Diffuse.r = parser.ReadFloat();
            material.m_Diffuse.g = parser.ReadFloat();
            material.m_Diffuse.b = parser.ReadFloat();
            parser.ReadToken( nrParser::TOKEN_LIST_END );
            material.m_Ambient.r = material.m_Diffuse.r;
            material.m_Ambient.g = material.m_Diffuse.g;
            material.m_Ambient.b = material.m_Diffuse.b;
        }
        else if ( parser.KeyMatches( key, "material" ) )
        {
            nrMaterial* m = nrMaterial::Parse( parser );
            material = *m;
            delete m;
        }
        else
        {
            parser.ParseError( "unknown key." );
        }
    }
    
    parser.ReadToken( nrParser::TOKEN_AGGREGATE_END );
    
    if ( parser.Error() )
    {
        return 0;
    }
    
    nrSurfaceSphere* sphere = new nrSurfaceSphere( center, radius, material );
    
    return sphere;
}

////////////////////////////////////////////////////////////////////////////
