////////////////////////////////////////////////////////////////////////////
//
// nrSurfaceTriangle.cpp
//
// A class for a triangle.
//
// Nate Robins, January 2002.
//
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Includes
////////////////////////////////////////////////////////////////////////////

#include "nrSurfaceTriangle.h"

#include "nrHit.h"
#include "nrInterval.h"
#include "nrLog.h"
#include "nrMath.h"
#include "nrParser.h"
#include "nrRay.h"

#include <assert.h>
#include <stdio.h>


////////////////////////////////////////////////////////////////////////////
// Public
////////////////////////////////////////////////////////////////////////////

static float epsilon = 0.0001f;

nrSurfaceTriangle::nrSurfaceTriangle( const nrVector3& a, const nrVector3& b, const nrVector3& c, const nrMaterial& material )
{
    m_A = a;
    m_B = b;
	m_C = c;
    
    if ( a.x < b.x && a.x < c.x )
    {
        m_Bound.m_Minimums.x = a.x;
    }
    else if ( b.x < a.x && b.x < c.x )
    {
        m_Bound.m_Minimums.x = b.x;
    }
    else
    {
        m_Bound.m_Minimums.x = c.x;
    }
    
    if ( a.y < b.y && a.y < c.y )
    {
        m_Bound.m_Minimums.y = a.y;
    }
    else if ( b.y < a.y && b.y < c.y )
    {
        m_Bound.m_Minimums.y = b.y;
    }
    else
    {
        m_Bound.m_Minimums.y = c.y;
    }
    
    if ( a.z < b.z && a.z < c.z )
    {
        m_Bound.m_Minimums.z = a.z;
    }
    else if ( b.z < a.z && b.z < c.z )
    {
        m_Bound.m_Minimums.z = b.z;
    }
    else
    {
        m_Bound.m_Minimums.z = c.z;
    }
    
    if ( a.x > b.x && a.x > c.x )
    {
        m_Bound.m_Maximums.x = a.x;
    }
    else if ( b.x > a.x && b.x > c.x )
    {
        m_Bound.m_Maximums.x = b.x;
    }
    else
    {
        m_Bound.m_Maximums.x = c.x;
    }
    
    if ( a.y > b.y && a.y > c.y )
    {
        m_Bound.m_Maximums.y = a.y;
    }
    else if ( b.y > a.y && b.y > c.y )
    {
        m_Bound.m_Maximums.y = b.y;
    }
    else
    {
        m_Bound.m_Maximums.y = c.y;
    }
    
    if ( a.z > b.z && a.z > c.z )
    {
        m_Bound.m_Maximums.z = a.z;
    }
    else if ( b.z > a.z && b.z > c.z )
    {
        m_Bound.m_Maximums.z = b.z;
    }
    else
    {
        m_Bound.m_Maximums.z = c.z;
    }
    
    // Make sure the bounding box has >~0 area
    float epsilon = 0.23f;
    if ( nrMath::Equal( m_Bound.m_Maximums.x, m_Bound.m_Minimums.x ) )
    {
        m_Bound.m_Minimums.x -= epsilon / 2.0f;
        m_Bound.m_Maximums.x += epsilon / 2.0f;
    }
    if ( nrMath::Equal( m_Bound.m_Maximums.y, m_Bound.m_Minimums.y ) )
    {
        m_Bound.m_Minimums.y -= epsilon / 2.0f;
        m_Bound.m_Maximums.y += epsilon / 2.0f;
    }
    if ( nrMath::Equal( m_Bound.m_Maximums.z, m_Bound.m_Minimums.z ) )
    {
        m_Bound.m_Minimums.z -= epsilon / 2.0f;
        m_Bound.m_Maximums.z += epsilon / 2.0f;
    }
    
    m_Material = material;
}

////////////////////////////////////////////////////////////////////////////

nrSurfaceTriangle::~nrSurfaceTriangle( void )
{
    //delete m_Material;
}

////////////////////////////////////////////////////////////////////////////

nrVector3 nrSurfaceTriangle::Normal( const nrVector3& point ) const
{
    return ( ( m_B - m_A ).Cross( m_C - m_A ) ).Unit();
}

////////////////////////////////////////////////////////////////////////////

bool nrSurfaceTriangle::Hit( const nrRay& ray, nrInterval& interval, nrHit& hit ) const
{
	float a = m_A.x - m_B.x;
	float b = m_A.y - m_B.y;
	float c = m_A.z - m_B.z;
	float d = m_A.x - m_C.x;
	float e = m_A.y - m_C.y;
	float f = m_A.z - m_C.z;
	float g = ray.d.x;
	float h = ray.d.y;
	float i = ray.d.z;
	float j = m_A.x - ray.o.x;
	float k = m_A.y - ray.o.y;
	float l = m_A.z - ray.o.z;
	
	float ei_minus_hf = e * i - h * f;
	float gf_minus_di = g * f - d * i;
	float dh_minus_eg = d * h - e * g;
	float ak_minus_jb = a * k - j * b;
	float jc_minus_al = j * c - a * l;
	float bl_minus_kc = b * l - k * c;
	
	float m = 1.0f / ( a * ei_minus_hf + b * gf_minus_di + c * dh_minus_eg );
	
	float beta = ( j * ei_minus_hf + k * gf_minus_di + l * dh_minus_eg ) * m;
	
	if ( beta >= 0 )
	{
		float gamma = ( i * ak_minus_jb + h * jc_minus_al + g * bl_minus_kc ) * m;
		
		if ( gamma >= 0 )
		{
			if ( ( beta + gamma ) <= 1 )
			{
				float t = -( f * ak_minus_jb + e * jc_minus_al + d * bl_minus_kc ) * m;
				
				if ( interval.Includes( t ) )
				{
                    hit = nrHit( this, t );
					return true;
				}
			}
		}
	}
	
    return false;
}

////////////////////////////////////////////////////////////////////////////

// !!! add checks for ensuring that at least a, b & c were read.

nrSurfaceTriangle* nrSurfaceTriangle::Parse( nrParser& parser )
{
    nrVector3  a, b, c;
    nrMaterial material;
    
    if ( parser.NextToken() == nrParser::TOKEN_STRING )
    {
        const char* name = parser.ReadString();
    }
    
    parser.ReadToken( nrParser::TOKEN_AGGREGATE_BEGIN );
    
    while ( parser.NextToken() == nrParser::TOKEN_KEY )
    {
        const char* key = parser.ReadKey();
        
        if ( parser.KeyMatches( key, "a" ) )
        {
            parser.ReadToken( nrParser::TOKEN_LIST_BEGIN );
            a.x = parser.ReadFloat();
            a.y = parser.ReadFloat();
            a.z = parser.ReadFloat();
            parser.ReadToken( nrParser::TOKEN_LIST_END );
        }
        else if ( parser.KeyMatches( key, "b" ) )
        {
            parser.ReadToken( nrParser::TOKEN_LIST_BEGIN );
            b.x = parser.ReadFloat();
            b.y = parser.ReadFloat();
            b.z = parser.ReadFloat();
            parser.ReadToken( nrParser::TOKEN_LIST_END );
        }
        else if ( parser.KeyMatches( key, "c" ) )
        {
            parser.ReadToken( nrParser::TOKEN_LIST_BEGIN );
            c.x = parser.ReadFloat();
            c.y = parser.ReadFloat();
            c.z = parser.ReadFloat();
            parser.ReadToken( nrParser::TOKEN_LIST_END );
        }
        else if ( parser.KeyMatches( key, "color" ) )
        {
            parser.ReadToken( nrParser::TOKEN_LIST_BEGIN );
            material.m_Diffuse.r = parser.ReadFloat();
            material.m_Diffuse.g = parser.ReadFloat();
            material.m_Diffuse.b = parser.ReadFloat();
            parser.ReadToken( nrParser::TOKEN_LIST_END );
            material.m_Ambient.r = material.m_Diffuse.r;
            material.m_Ambient.g = material.m_Diffuse.g;
            material.m_Ambient.b = material.m_Diffuse.b;
        }
        else if ( parser.KeyMatches( key, "material" ) )
        {
            nrMaterial* m = nrMaterial::Parse( parser );
            material = *m;
            delete m;
        }
        else
        {
            parser.ParseError( "unknown key." );
        }
    }
    
    parser.ReadToken( nrParser::TOKEN_AGGREGATE_END );
    
    if ( parser.Error() )
    {
        return 0;
    }
    
    nrSurfaceTriangle* triangle = new nrSurfaceTriangle( a, b, c, material );
    
    return triangle;
}

////////////////////////////////////////////////////////////////////////////
