////////////////////////////////////////////////////////////////////////////
//
// nrSurfaceSphere.cpp
//
// A class for a sphere.
//
// Nate Robins, January 2002.
//
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Includes
////////////////////////////////////////////////////////////////////////////

#include "nrSurfaceSphere.h"

#include "nrHit.h"
#include "nrInterval.h"
#include "nrLog.h"
#include "nrMath.h"
#include "nrRay.h"

#include <assert.h>
#include <stdio.h>


////////////////////////////////////////////////////////////////////////////
// Public
////////////////////////////////////////////////////////////////////////////

nrSurfaceSphere::nrSurfaceSphere( const nrVector3& center, float radius, const nrMaterial& material )
{
    m_Center        = center;
    m_Radius        = radius;
    m_Material      = material;
}

////////////////////////////////////////////////////////////////////////////

nrSurfaceSphere::~nrSurfaceSphere( void )
{
}

////////////////////////////////////////////////////////////////////////////

nrVector3 nrSurfaceSphere::Normal( const nrVector3& point ) const
{
    return ( point - m_Center ).Unit();
}

////////////////////////////////////////////////////////////////////////////


inline __int64 CpuClock( void )
{
	__int64 c;

	__asm
	{
		rdtsc;
		mov	dword ptr c+0, eax;
		mov	dword ptr c+4, edx;
	}

	return c;
}

float epsilon = 0.0001f;

__int64 allclocks = 0;
int allhits = 0;

void ShowClockCycles( )
{
	printf( "Ave. Clock Cycle = %d\n", (allclocks / allhits) );
}

bool nrSurfaceSphere::Hit( const nrRay& ray, nrInterval& interval, nrHit& hit ) const
{
	__int64 clock1;
	__int64 clock2;

	allhits++;

	clock1 = CpuClock();

    const nrVector3& d = ray.d;
    const nrVector3 co = ray.o - m_Center;
    
    float A = d.Dot( d );       assert( ! nrMath::Equal( A, 0.0f ) );
    float B = 2 * d.Dot( co );
    float C = co.Dot( co ) - m_Radius * m_Radius;
    float discriminant = B * B - 4 * A * C;
    
    if ( discriminant >= epsilon )
    {
		 float aa = 1.0f / ( 2 * A );
        float s = nrMath::Sqrt( discriminant );
        
        //float t = ( -B - s ) / ( 2 * A );
        float t = ( -B - s ) * aa;
        if ( interval.Includes( t ) ) 
        {
            hit.t = t;
            hit.m_Surface = this;

				clock2 = CpuClock();
				allclocks += (clock2 - clock1);
            return true;
        }
        
        //t = ( -B + s ) / ( 2 * A );
        t = ( -B + s ) * aa;
        if ( interval.Includes( t ) )
        {
            hit.t = t;
            hit.m_Surface = this;
				clock2 = CpuClock();
				allclocks += (clock2 - clock1);
            return true;
        }
    }
    
	 clock2 = CpuClock();
	 allclocks += (clock2 - clock1);
    return false;
}

////////////////////////////////////////////////////////////////////////////

nrSurfaceSphere* nrSurfaceSphere::NewFromString( const char* string )
{
    nrVector3  center;
    float      radius;
    nrMaterial material;

    const nrColor& diffuse = material.diffuse;

    if ( sscanf( string, "%f %f %f %f %f %f %f", &center.x, &center.y, &center.z, &radius, &diffuse.r, &diffuse.g, &diffuse.b ) != 7 )
    {
        return 0;
    }
    
    nrSurfaceSphere* sphere = new nrSurfaceSphere( center, radius, material );
    if ( sphere == 0 )
    {
        g_Log.Write( "nrSurfaceSphere::NewFromString(): Out of memory.\n" );
    }
    
    return sphere;
}

////////////////////////////////////////////////////////////////////////////
