////////////////////////////////////////////////////////////////////////////
//
// raynOptions.cpp
//
// Class for options for Nate's Ray Tracer.
//
// Nate Robins, January 2002
//
////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////
// Includes
////////////////////////////////////////////////////////////////////////////

#include "raynOptions.h"

#include "nrCommandLine.h"
#include "nrLog.h"

#include <string.h>
#include <stdlib.h>


////////////////////////////////////////////////////////////////////////////
// Public
////////////////////////////////////////////////////////////////////////////

raynOptions::raynOptions()
{
    width      = 256;
    height     = 256;
    shadows    = true;
    bvh        = true;
    
    output = new char[ strlen( "out.ppm" ) + 1 ];
    strcpy( output, "out.ppm" );
    
    scene = new char[ strlen( "scene.txt" ) + 1 ];
    strcpy( scene, "scene.txt" );
}

////////////////////////////////////////////////////////////////////////////

raynOptions::~raynOptions()
{
    delete [] output;
    delete [] scene;
}

////////////////////////////////////////////////////////////////////////////

bool raynOptions::Parse( int argc, const char** argv )
{
    nrCommandLine options;

    if ( ! options.Create( argc, argv ) )
    {
        Usage( argv[ 0 ] );  
        return false;
    }
    
    if ( options.NumOptions() == 0 )
    {
        g_Log.Write( "\nScene file argument required.\n" );
        Usage( argv[ 0 ] );
        return false;
    }
    
    for ( int i = 0; i < options.NumOptions(); i++ )
    {
        const char* option   = options.Option( i );
        const char* argument = options.Argument( i );
        
        if ( strcmp( "size", option ) == 0 )
        {
            if ( argument == 0 )
            {
                g_Log.Write( "\nOption -%s requires an argument.\n", option );
                Usage( argv[ 0 ] );
                return false;
            }
            
            int w, h;
            sscanf( argument, "%d,%d", &w, &h );
            
            width = w;
            height = h;
        }
        else if ( strcmp( "nx", option ) == 0 )
        {
            if ( argument == 0 )
            {
                g_Log.Write( "\nOption -%s requires an argument.\n", option );
                Usage( argv[ 0 ] );
                return false;
            }
            
            int w;
            sscanf( argument, "%d", &w );
            
            width = w;
        }
        else if ( strcmp( "ny", option ) == 0 )
        {
            if ( argument == 0 )
            {
                g_Log.Write( "\nOption -%s requires an argument.\n", option );
                Usage( argv[ 0 ] );
                return false;
            }
            
            int h;
            sscanf( argument, "%d", &h );
            
            height = h;
        }
        else if ( strcmp( "noshadows", option ) == 0 )
        {
            shadows = false;
        }
        else if ( strcmp( "nobvh", option ) == 0 )
        {
            bvh = false;
        }
        else if ( strcmp( "output", option ) == 0 ||
            strcmp( "out",    option ) == 0 )
        {
            if ( argument == 0 )
            {
                g_Log.Write( "\nOption -%s requires an argument.\n", option );
                Usage( argv[ 0 ] );
                return false;
            }
            
            delete [] output;
            output = new char[ strlen( argument ) + 1 ];
            strcpy( output, argument );
        }
        else if ( strcmp( "scene", option ) == 0 ||
            strcmp( "model", option ) == 0 )
        {
            if ( argument == 0 )
            {
                g_Log.Write( "\nOption -%s requires an argument.\n", option );
                Usage( argv[ 0 ] );
                return false;
            }
            
            delete [] scene;
            scene = new char[ strlen( argument ) + 1 ];
            strcpy( scene, argument );
        }
        else
        {
            g_Log.Write( "\nUnknown option -%s.\n", option );
            Usage( argv[ 0 ] );
            return false;
        }
    }
    
    return true;
}

////////////////////////////////////////////////////////////////////////////

void raynOptions::Usage( const char* program_name )
{
#ifdef WIN32
    {
        // Under Windows, the program_name is the full path.  Cut that down to 
        // just the basename.
        char drive[ _MAX_DRIVE ];
        char directory[ _MAX_DIR ];
        char extension[ _MAX_EXT ];
        static char filename[ _MAX_FNAME ];
        
        _splitpath( program_name, drive, directory, filename, extension );
        
        program_name = filename;
    }
#endif
    
    g_Log.Write( "\n" );
    g_Log.Write( "%s [@response] [options]\n", program_name );
    g_Log.Write( "\n" );
    g_Log.Write( "  options:\n" );
    g_Log.Write( "    -scene      <scene_file>       file containing scene description (string).\n" );
    g_Log.Write( "    -output     <output_file>      output image filename (string).\n" );
    g_Log.Write( "    -size       <width,height>     size of output image (int,int).\n" );
    g_Log.Write( "    -noshadows                     don't generate shadow rays.\n" );
    g_Log.Write( "    -nobvh                         don't generate a bounding volume hierarchy.\n" );
    g_Log.Write( "\n" );
    g_Log.Write( "  @reponse:\n" );
    g_Log.Write( "    Specifies a file from which options are to be read,\n" );
    g_Log.Write( "    one option per line.  Options which are specified on\n" );
    g_Log.Write( "    the command line override those in the response file.\n" );
    g_Log.Write( "\n" );
    g_Log.Write( "\n" );
}

////////////////////////////////////////////////////////////////////////////
